/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import org.sablecc.sablecc.analysis.AnalysisAdapter;
import org.sablecc.sablecc.node.ATokenDef;
import org.sablecc.sablecc.node.Start;

public class GenLexer extends AnalysisAdapter
{
  private MacroExpander macros;
  private ResolveIds ids;
  private File pkgDir;
  private AcceptStates[] acceptStatesArray;
  private Transitions transitions;

  public GenLexer(ResolveIds ids)
  {
    this.ids = ids;

    try
    {
      macros = new MacroExpander(
                 new InputStreamReader(
                   getClass().getResourceAsStream("lexer.txt")));
    }
    catch(IOException e)
    {
      throw new RuntimeException("unable to open lexer.txt.", e);
    }

    pkgDir = new File(ids.pkgDir, "lexer");

    if(!pkgDir.exists())
    {
      if(!pkgDir.mkdir())
      {
        throw new RuntimeException("Unable to create " + pkgDir.getAbsolutePath());
      }
    }
  }

  @Override
  public void caseStart(Start tree)
  {
    String[] names;
    int numStates = Math.max(1, ids.stateList.size());

    acceptStatesArray = new AcceptStates[numStates];
    names = new String[numStates];

    if(ids.stateList.size() == 0)
    {
      names[0] = "INITIAL";
    }
    else
    {
      Iterator<String> iter = ids.stateList.iterator();
      for(int i = 0; i < numStates; i++)
      {
        names[i] = iter.next();
      }
    }

    for(int i = 0; i < numStates; i++)
    {
      System.out.println(" State: " + names[i]);

      System.out.println(" - Constructing NFA.");
      ConstructNFA nfaConstructor = new ConstructNFA(ids, names[i]);
      tree.apply(nfaConstructor);
      System.out.println();

      NFA nfa = nfaConstructor.getNFA();
      nfaConstructor = null;

      System.out.println(" - Constructing DFA.");
      DFA dfa = new DFA(nfa);
      System.out.println();

      System.out.println(" - resolving ACCEPT states.");
      acceptStatesArray[i] = new AcceptStates(dfa, ids, names[i]);
      tree.apply(acceptStatesArray[i]);
    }

    transitions = new Transitions();
    tree.apply(transitions);

    createLexerException();
    createLexer();
  }

  private void createLexerException()
  {
    try(BufferedWriter file = new BufferedWriter(new FileWriter(new File(pkgDir, "LexerException.java"))))
    {
      macros.apply(file, "LexerException", new String[] {ids.pkgNameDot});
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "LexerException.java").getAbsolutePath(), e);
    }
  }

  private void createLexer()
  {
    try(BufferedWriter file = new BufferedWriter(new FileWriter(new File(pkgDir, "Lexer.java"))))
    {
      String startState = "INITIAL";
      if(ids.stateList.size() > 0)
      {
        startState = ids.stateList.getFirst();
      }

      macros.apply(file, "LexerHeader", new String[] {ids.pkgNameDot, startState});

      for(ListIterator<String> i = ids.tokenList.listIterator(); i.hasNext();)
      {
        String name = i.next();
        ATokenDef node = ids.tokens.get(name);
        boolean fixed = ids.fixedTokens.get(node);

        if(fixed)
        {
          macros.apply(file, "LexerFixedToken",
                       new String[] {"" + i.previousIndex(), name});
        }
        else
        {
          macros.apply(file, "LexerVariableToken",
                       new String[] {"" + i.previousIndex(), name});
        }

        Map<String, String> map = transitions.tokenStates.get(node);
        if(map.size() > 0)
        {
          macros.apply(file, "TokenSwitchHeader", null);

          for(Map.Entry<String, String> entry : map.entrySet())
          {
            macros.apply(file, "TokenCase",
                         new String[] {ids.stateList.indexOf(entry.getKey()) + "", entry.getValue()});
          }

          macros.apply(file, "TokenSwitchTail", null);
        }

        macros.apply(file, "LexerTokenTail", null);
      }

      macros.apply(file, "LexerBody1");

      for(ListIterator<String> i = ids.tokenList.listIterator(); i.hasNext();)
      {
        String name = i.next();
        ATokenDef node = ids.tokens.get(name);
        boolean fixed = ids.fixedTokens.get(node);

        if(fixed)
        {
          macros.apply(file, "LexerNewFixedToken",
                       new String[] {"" + i.previousIndex(), name});
        }
        else
        {
          macros.apply(file, "LexerNewVariableToken",
                       new String[] {"" + i.previousIndex(), name});
        }
      }

      macros.apply(file, "LexerBody2");

      DataOutputStream out = new DataOutputStream(
                               new BufferedOutputStream(
                                 new FileOutputStream(
                                   new File(pkgDir, "lexer.dat"))));

      out.writeInt(acceptStatesArray.length);
      for(int accSt = 0; accSt < acceptStatesArray.length; accSt++)
      {
        DFA dfa = acceptStatesArray[accSt].dfa;

        file.write("        { // " + acceptStatesArray[accSt].stateName + System.getProperty("line.separator"));
        List<List<int[]>> outerArray = new ArrayList<>();

        for(DFA.State state : dfa.states)
        {
          List<int[]> innerArray = new ArrayList<>();

          file.write("            {");

          for(DFA.Transition transition : state.transitions)
          {
            file.write("{" + ((int) transition.interval().start) + ", " +
                       ((int) transition.interval().end) + ", " +
                       transition.destination + "}, ");

            innerArray.add(new int[] {
                                    ((int) transition.interval().start),
                                    ((int) transition.interval().end),
                                    transition.destination});
          }

          file.write("}," + System.getProperty("line.separator"));

          outerArray.add(innerArray);
        }
        file.write("        }" + System.getProperty("line.separator"));

        out.writeInt(outerArray.size());
        for(List<int[]> innerArray : outerArray)
        {
          out.writeInt(innerArray.size());
          for(int[] array : innerArray)
          {
            for(int i = 0; i < 3; i++)
            {
              out.writeInt(array[i]);
            }
          }
        }
      }

      macros.apply(file, "LexerAcceptHeader");

      final int stateNumber = acceptStatesArray.length;

      List<List<Integer>> outerArray = new ArrayList<>();

      for(int i = 0; i < stateNumber; i++)
      {
        DFA dfa = acceptStatesArray[i].dfa;
        List<Integer> innerArray = new ArrayList<>();

        file.write("        // " + acceptStatesArray[i].stateName + System.getProperty("line.separator"));
        file.write("        {");

        for(DFA.State state : dfa.states)
        {
          file.write(state.accept + ", ");
          innerArray.add(state.accept);
        }

        file.write("}," + System.getProperty("line.separator"));

        outerArray.add(innerArray);
      }

      out.writeInt(outerArray.size());
      for(List<Integer> innerArray : outerArray)
      {
        out.writeInt(innerArray.size());
        for(int i : innerArray)
        {
          out.writeInt(i);
        }
      }
      out.close();

      file.write(System.getProperty("line.separator"));

      macros.apply(file, "LexerAcceptTail");

      macros.apply(file, "LexerStateHeader");

      if(ids.stateList.size() > 0)
      {
        for(ListIterator<String> i = ids.stateList.listIterator(); i.hasNext();)
        {
          String s = i.next();

          macros.apply(file, "LexerStateBody",
                       new String[] {s, "" + i.previousIndex()});
        }
      }
      else
      {
        macros.apply(file, "LexerStateBody",
                     new String[] {"INITIAL", "" + 0});
      }

      macros.apply(file, "LexerStateTail");

      macros.apply(file, "LexerTail");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Lexer.java").getAbsolutePath(), e);
    }
  }
}
